@@ -1,24 +1,22 @@
 use crate::common::init_special_chars_once;
 use crate::complete::complete_invalidate_path;
 use crate::env::{DEFAULT_READ_BYTE_LIMIT, READ_BYTE_LIMIT};
-use crate::env::{EnvMode, EnvStack, Environment, setenv_lock, unsetenv_lock};
+use crate::env::{EnvMode, EnvStack, Environment as _, setenv_lock, unsetenv_lock};
 use crate::flog::flog;
+use crate::function;
 use crate::input_common::{update_wait_on_escape_ms, update_wait_on_sequence_key_ms};
 use crate::locale::{invalidate_numeric_locale, set_libc_locales};
 use crate::prelude::*;
 use crate::reader::{
     reader_change_cursor_end_mode, reader_change_cursor_selection_mode, reader_change_history,
-    reader_schedule_prompt_repaint, reader_set_autosuggestion_enabled, reader_set_transient_prompt,
-};
-use crate::screen::{
-    IS_DUMB, LAYOUT_CACHE_SHARED, ONLY_GRAYSCALE, screen_set_midnight_commander_hack,
+    reader_current_data, reader_schedule_prompt_repaint, reader_set_autosuggestion_enabled,
+    reader_set_transient_prompt,
 };
+use crate::screen::{IS_DUMB, ONLY_GRAYSCALE, screen_set_midnight_commander_hack};
 use crate::terminal::ColorSupport;
-use crate::terminal::use_terminfo;
 use crate::tty_handoff::xtversion;
-use crate::wcstringutil::string_prefixes_string;
 use crate::wutil::fish_wcstoi;
-use crate::{function, terminal};
+use fish_wcstringutil::{bool_from_string, string_prefixes_string};
 use std::collections::HashMap;
 use std::sync::atomic::{AtomicBool, Ordering};
 
@@ -34,11 +32,6 @@ const LOCALE_VARIABLES: [&wstr; 7] = [
     L!("LOCPATH"),
 ];
 
-#[rustfmt::skip]
-const CURSES_VARIABLES: [&wstr; 3] = [
-    L!("TERM"), L!("TERMINFO"), L!("TERMINFO_DIRS")
-];
-
 /// Whether to use `posix_spawn()` when possible.
 static USE_POSIX_SPAWN: AtomicBool = AtomicBool::new(false);
 
@@ -47,55 +40,65 @@ static VAR_DISPATCH_TABLE: once_cell::sync::Lazy<VarDispatchTable> =
     once_cell::sync::Lazy::new(|| {
         let mut table = VarDispatchTable::default();
 
-        for name in LOCALE_VARIABLES {
-            table.add_anon(name, handle_locale_change);
+        macro_rules! vars {
+            ( $f:ident ) => {
+                |vars: &EnvStack, _suppress_repaint: bool| $f(vars)
+            };
         }
 
-        for name in CURSES_VARIABLES {
-            table.add_anon(name, handle_term_change);
+        for name in LOCALE_VARIABLES {
+            table.add_anon(name, vars!(handle_locale_change));
         }
 
+        table.add_anon(L!("TERM"), handle_term_change);
+
         table.add(L!("TZ"), handle_tz_change);
         table.add_anon(L!("COLORTERM"), handle_fish_term_change);
         table.add_anon(L!("fish_term256"), handle_fish_term_change);
         table.add_anon(L!("fish_term24bit"), handle_fish_term_change);
-        table.add_anon(L!("fish_escape_delay_ms"), update_wait_on_escape_ms);
+        table.add_anon(L!("fish_escape_delay_ms"), vars!(update_wait_on_escape_ms));
         table.add_anon(
             L!("fish_sequence_key_delay_ms"),
-            update_wait_on_sequence_key_ms,
+            vars!(update_wait_on_sequence_key_ms),
         );
-        table.add_anon(L!("fish_emoji_width"), guess_emoji_width);
-        table.add_anon(L!("fish_ambiguous_width"), handle_change_ambiguous_width);
-        table.add_anon(L!("LINES"), handle_term_size_change);
-        table.add_anon(L!("COLUMNS"), handle_term_size_change);
-        table.add_anon(L!("fish_complete_path"), handle_complete_path_change);
-        table.add_anon(L!("fish_function_path"), handle_function_path_change);
-        table.add_anon(L!("fish_read_limit"), handle_read_limit_change);
-        table.add_anon(L!("fish_history"), handle_fish_history_change);
+        table.add_anon(L!("fish_emoji_width"), vars!(guess_emoji_width));
+        table.add_anon(
+            L!("fish_ambiguous_width"),
+            vars!(handle_change_ambiguous_width),
+        );
+        table.add_anon(L!("LINES"), vars!(handle_term_size_change));
+        table.add_anon(L!("COLUMNS"), vars!(handle_term_size_change));
+        table.add_anon(L!("fish_complete_path"), vars!(handle_complete_path_change));
+        table.add_anon(L!("fish_function_path"), vars!(handle_function_path_change));
+        table.add_anon(L!("fish_read_limit"), vars!(handle_read_limit_change));
+        table.add_anon(L!("fish_history"), vars!(handle_fish_history_change));
         table.add_anon(
             L!("fish_autosuggestion_enabled"),
-            handle_autosuggestion_change,
+            vars!(handle_autosuggestion_change),
+        );
+        table.add_anon(
+            L!("fish_transient_prompt"),
+            vars!(handle_transient_prompt_change),
         );
-        table.add_anon(L!("fish_transient_prompt"), handle_transient_prompt_change);
         table.add_anon(
             L!("fish_use_posix_spawn"),
-            handle_fish_use_posix_spawn_change,
+            vars!(handle_fish_use_posix_spawn_change),
         );
-        table.add_anon(L!("fish_trace"), handle_fish_trace);
+        table.add_anon(L!("fish_trace"), vars!(handle_fish_trace));
         table.add_anon(
             L!("fish_cursor_selection_mode"),
-            handle_fish_cursor_selection_mode_change,
+            vars!(handle_fish_cursor_selection_mode_change),
         );
         table.add_anon(
             L!("fish_cursor_end_mode"),
-            handle_fish_cursor_end_mode_change,
+            vars!(handle_fish_cursor_end_mode_change),
         );
 
         table
     });
 
 type NamedEnvCallback = fn(name: &wstr, env: &EnvStack);
-type AnonEnvCallback = fn(env: &EnvStack);
+type AnonEnvCallback = fn(env: &EnvStack, suppress_repaint: bool);
 
 enum EnvCallback {
     Named(NamedEnvCallback),
@@ -120,10 +123,10 @@ impl VarDispatchTable {
         assert!(prev.is_none(), "Already observing {}", name);
     }
 
-    pub fn dispatch(&self, key: &wstr, vars: &EnvStack) {
+    pub fn dispatch(&self, key: &wstr, vars: &EnvStack, suppress_repaint: bool) {
         match self.table.get(key) {
             Some(EnvCallback::Named(named)) => (named)(key, vars),
-            Some(EnvCallback::Anon(anon)) => (anon)(vars),
+            Some(EnvCallback::Anon(anon)) => (anon)(vars, suppress_repaint),
             None => (),
         }
     }
@@ -206,26 +209,40 @@ pub fn guess_emoji_width(vars: &EnvStack) {
     }
 }
 
+pub struct VarChangeMilieu {
+    pub is_repainting: bool,
+    pub global_or_universal: bool,
+}
+
 /// React to modifying the given variable.
-pub fn env_dispatch_var_change(key: &wstr, vars: &EnvStack) {
+pub fn env_dispatch_var_change(milieu: VarChangeMilieu, key: &wstr, vars: &EnvStack) {
     use once_cell::sync::Lazy;
 
+    let suppress_repaint = milieu.is_repainting || !milieu.global_or_universal;
+
     // We want to ignore variable changes until the dispatch table is explicitly initialized.
+    // TODO(MSRV>=1.94): Use std::sync::LazyLock. (LazyLock::get is stabilized in Rust 1.94)
     if let Some(dispatch_table) = Lazy::get(&VAR_DISPATCH_TABLE) {
-        dispatch_table.dispatch(key, vars);
+        dispatch_table.dispatch(key, vars, suppress_repaint);
     }
 
-    if string_prefixes_string(L!("fish_color_"), key)
-        // TODO Don't re-exec prompt when only pager color changed.
-        || string_prefixes_string(L!("fish_pager_color_"), key)
-    {
-        reader_schedule_prompt_repaint();
+    if !suppress_repaint {
+        if let Some(data) = reader_current_data() {
+            if string_prefixes_string(L!("fish_color_"), key) || {
+                // TODO Don't re-exec prompt when only pager color changed.
+                string_prefixes_string(L!("fish_pager_color_"), key)
+            } {
+                data.schedule_prompt_repaint();
+            }
+        }
     }
 }
 
-fn handle_fish_term_change(vars: &EnvStack) {
+fn handle_fish_term_change(vars: &EnvStack, suppress_repaint: bool) {
     update_fish_color_support(vars);
-    reader_schedule_prompt_repaint();
+    if !suppress_repaint {
+        reader_schedule_prompt_repaint();
+    }
 }
 
 fn handle_change_ambiguous_width(vars: &EnvStack) {
@@ -296,7 +313,7 @@ fn handle_function_path_change(_: &EnvStack) {
 }
 
 fn handle_complete_path_change(_: &EnvStack) {
-    complete_invalidate_path()
+    complete_invalidate_path();
 }
 
 fn handle_tz_change(var_name: &wstr, vars: &EnvStack) {
@@ -307,11 +324,12 @@ fn handle_locale_change(vars: &EnvStack) {
     init_locale(vars);
 }
 
-fn handle_term_change(vars: &EnvStack) {
+fn handle_term_change(vars: &EnvStack, suppress_repaint: bool) {
     guess_emoji_width(vars);
     init_terminal(vars);
-    read_terminfo_database(vars);
-    reader_schedule_prompt_repaint();
+    if !suppress_repaint {
+        reader_schedule_prompt_repaint();
+    }
 }
 
 fn handle_fish_use_posix_spawn_change(vars: &EnvStack) {
@@ -319,8 +337,7 @@ fn handle_fish_use_posix_spawn_change(vars: &EnvStack) {
     if !cfg!(have_posix_spawn) {
         USE_POSIX_SPAWN.store(false, Ordering::Relaxed);
     } else if let Some(var) = vars.get(L!("fish_use_posix_spawn")) {
-        let use_posix_spawn =
-            var.is_empty() || crate::wcstringutil::bool_from_string(&var.as_string());
+        let use_posix_spawn = var.is_empty() || bool_from_string(&var.as_string());
         USE_POSIX_SPAWN.store(use_posix_spawn, Ordering::Relaxed);
     } else {
         USE_POSIX_SPAWN.store(true, Ordering::Relaxed);
@@ -387,14 +404,14 @@ fn run_inits(vars: &EnvStack) {
 /// Updates our idea of whether we support term256 and term24bit (see issue #10222).
 fn update_fish_color_support(vars: &EnvStack) {
     // Detect or infer term256 support. If fish_term256 is set, we respect it. Otherwise, infer it
-    // from $TERM or use terminfo.
+    // from $TERM.
 
     let term = vars.get_unless_empty(L!("TERM"));
     let term = term.as_ref().map_or(L!(""), |term| &term.as_list()[0]);
     let is_xterm_16color = term == "xterm-16color";
 
     let supports_256color = if let Some(fish_term256) = vars.get(L!("fish_term256")) {
-        let ok = crate::wcstringutil::bool_from_string(&fish_term256.as_string());
+        let ok = bool_from_string(&fish_term256.as_string());
         flog!(
             term_support,
             "256-color support determined by $fish_term256:",
@@ -409,7 +426,7 @@ fn update_fish_color_support(vars: &EnvStack) {
     #[allow(unused_parens)]
     if let Some(fish_term24bit) = vars.get(L!("fish_term24bit")).map(|v| v.as_string()) {
         // $fish_term24bit
-        supports_24bit = crate::wcstringutil::bool_from_string(&fish_term24bit);
+        supports_24bit = bool_from_string(&fish_term24bit);
         flog!(
             term_support,
             "$fish_term24bit preference: 24-bit color",
@@ -486,32 +503,6 @@ fn init_terminal(vars: &EnvStack) {
     update_fish_color_support(vars);
 }
 
-pub fn read_terminfo_database(vars: &EnvStack) {
-    if !use_terminfo() {
-        return;
-    }
-
-    // The current process' environment needs to be modified because the terminfo crate will
-    // read these variables
-    for var_name in CURSES_VARIABLES {
-        if let Some(value) = vars
-            .getf_unless_empty(var_name, EnvMode::EXPORT)
-            .map(|v| v.as_string())
-        {
-            flog!(term_support, "curses var", var_name, "=", value);
-            setenv_lock(var_name, &value, true);
-        } else {
-            flog!(term_support, "curses var", var_name, "is missing or empty");
-            unsetenv_lock(var_name);
-        }
-    }
-
-    terminal::setup();
-
-    // Invalidate the cached escape sequences since they may no longer be valid.
-    LAYOUT_CACHE_SHARED.lock().unwrap().clear();
-}
-
 /// Initialize the locale subsystem
 fn init_locale(vars: &EnvStack) {
     let _guard = crate::locale::LOCALE_LOCK.lock().unwrap();
