@@ -22,14 +22,15 @@ use crate::{
         rewrite_via_temporary_file,
     },
     threads::ThreadPool,
-    wcstringutil::trim,
 };
+use fish_wcstringutil::{subsequence_in_string, trim};
+use fish_widestring::subslice_position;
 use std::{
     borrow::Cow,
     collections::{BTreeMap, HashMap, HashSet},
-    ffi::CString,
+    ffi::{CStr, CString},
     fs::File,
-    io::{BufRead, Read, Write},
+    io::{BufRead, BufWriter, Read as _, Write as _},
     mem::MaybeUninit,
     num::NonZeroUsize,
     ops::ControlFlow,
@@ -40,27 +41,27 @@ use std::{
 use bitflags::bitflags;
 use lru::LruCache;
 use nix::{fcntl::OFlag, sys::stat::Mode};
-use rand::Rng;
+use rand::Rng as _;
 
 use crate::{
-    ast::{self, Kind, Node},
+    ast::{self, Kind, Node as _},
     common::{CancelChecker, UnescapeStringStyle, bytes2wcstring, unescape_string, valid_var_name},
     env::{EnvMode, EnvStack, Environment},
     expand::{ExpandFlags, expand_one},
     fds::wopen_cloexec,
     flog::{flog, flogf},
     fs::fsync,
-    history::file::{HistoryFile, RawHistoryFile, append_history_item_to_buffer},
+    highlight::highlight_and_colorize,
+    history::file::{HistoryFile, RawHistoryFile},
     io::IoStreams,
     localization::wgettext_fmt,
     operation_context::{EXPANSION_LIMIT_BACKGROUND, OperationContext},
     parse_constants::{ParseTreeFlags, StatementDecoration},
-    parse_util::{parse_util_detect_errors, parse_util_unescape_wildcards},
+    parse_util::{detect_parse_errors, unescape_wildcards},
+    parser::Parser,
     path::{path_get_config, path_get_data, path_is_valid},
     prelude::*,
     threads::assert_is_background_thread,
-    util::find_subslice,
-    wcstringutil::subsequence_in_string,
     wildcard::{ANY_STRING, wildcard_match},
     wutil::{FileId, INVALID_FILE_ID, file_id_for_file, wrealpath, wstat, wunlink},
 };
@@ -81,8 +82,6 @@ pub enum SearchType {
     PrefixGlob,
     /// Search for commands containing the given string as a subsequence
     ContainsSubsequence,
-    /// Matches everything.
-    MatchEverything,
 }
 
 /// Ways that a history item may be written to disk (or omitted).
@@ -109,16 +108,6 @@ const DFLT_FISH_HISTORY_SESSION_ID: &wstr = L!("fish");
 
 pub const VACUUM_FREQUENCY: usize = 25;
 
-/// Output the contents `buffer` to `file` and clear the `buffer`.
-fn flush_to_file(buffer: &mut Vec<u8>, file: &mut File, min_size: usize) -> std::io::Result<()> {
-    if buffer.is_empty() || buffer.len() < min_size {
-        return Ok(());
-    }
-    file.write_all(buffer)?;
-    buffer.clear();
-    Ok(())
-}
-
 struct TimeProfiler {
     what: &'static str,
     start: SystemTime,
@@ -143,9 +132,9 @@ impl Drop for TimeProfiler {
                 self.what,
                 ms as u64, // todo!("remove cast")
                 ns as u32
-            )
+            );
         } else {
-            flogf!(profile_history, "%s: ??? ms", self.what)
+            flogf!(profile_history, "%s: ??? ms", self.what);
         }
     }
 }
@@ -227,7 +216,7 @@ impl HistoryItem {
         match typ {
             SearchType::Exact => term == *content_to_match,
             SearchType::Contains => {
-                find_subslice(term.as_slice(), content_to_match.as_slice()).is_some()
+                subslice_position(content_to_match.as_slice(), term.as_slice()).is_some()
             }
             SearchType::Prefix => content_to_match.as_slice().starts_with(term.as_slice()),
             SearchType::LinePrefix => content_to_match
@@ -235,7 +224,7 @@ impl HistoryItem {
                 .split(|&c| c == '\n')
                 .any(|line| line.starts_with(term.as_char_slice())),
             SearchType::ContainsGlob => {
-                let mut pat = parse_util_unescape_wildcards(term);
+                let mut pat = unescape_wildcards(term);
                 if !pat.starts_with(ANY_STRING) {
                     pat.insert(0, ANY_STRING);
                 }
@@ -245,14 +234,13 @@ impl HistoryItem {
                 wildcard_match(content_to_match.as_ref(), &pat, false)
             }
             SearchType::PrefixGlob => {
-                let mut pat = parse_util_unescape_wildcards(term);
+                let mut pat = unescape_wildcards(term);
                 if !pat.ends_with(ANY_STRING) {
                     pat.push(ANY_STRING);
                 }
                 wildcard_match(content_to_match.as_ref(), &pat, false)
             }
             SearchType::ContainsSubsequence => subsequence_in_string(term, &content_to_match),
-            SearchType::MatchEverything => true,
         }
     }
 
@@ -289,7 +277,7 @@ impl HistoryItem {
         // Ok, merge this item.
         self.creation_timestamp = self.creation_timestamp.max(item.creation_timestamp);
         if self.required_paths.len() < item.required_paths.len() {
-            self.required_paths = item.required_paths.clone();
+            self.required_paths.clone_from(&item.required_paths);
         }
         true
     }
@@ -297,9 +285,19 @@ impl HistoryItem {
 
 static HISTORIES: Mutex<BTreeMap<WString, Arc<History>>> = Mutex::new(BTreeMap::new());
 
+/// When deleting, whether the deletion should be only for this session or for all sessions.
+#[derive(Clone, Copy, PartialEq, Eq)]
+enum DeletionScope {
+    SessionOnly,
+    AllSessions,
+}
+
 struct HistoryImpl {
     /// The name of this list. Used for picking a suitable filename and for switching modes.
     name: WString,
+    /// Optional custom directory for the history file. If None, uses path_get_data().
+    /// Primarily for testing.
+    custom_directory: Option<WString>,
     /// New items. Note that these are NOT discarded on save. We need to keep these around so we can
     /// distinguish between items in our history and items in the history of other shells that were
     /// started after we were started.
@@ -311,10 +309,8 @@ struct HistoryImpl {
     has_pending_item: bool, // false
     /// Whether we should disable saving to the file for a time.
     disable_automatic_save_counter: u32, // 0
-    /// Deleted item contents.
-    /// Boolean describes if it should be deleted only in this session or in all
-    /// (used in deduplication).
-    deleted_items: HashMap<WString, bool>,
+    /// Deleted item contents, and the scope of the deletion.
+    deleted_items: HashMap<WString, DeletionScope>,
     /// The history file contents.
     file_contents: Option<HistoryFile>,
     /// The file ID of the history file.
@@ -340,7 +336,11 @@ impl HistoryImpl {
             return Ok(None);
         }
 
-        let Some(mut path) = path_get_data() else {
+        let mut path = if let Some(custom_dir) = &self.custom_directory {
+            custom_dir.clone()
+        } else if let Some(data_path) = path_get_data() {
+            data_path
+        } else {
             return Err(std::io::Error::new(
                 std::io::ErrorKind::NotFound,
                 "Error obtaining data directory. This is a manually constructed error which does not indicate why this happened.",
@@ -350,7 +350,11 @@ impl HistoryImpl {
         path.push('/');
         path.push_utfstr(&self.name);
         path.push_utfstr(L!("_history"));
-        if let Some(canonicalized_path) = wrealpath(&path) {
+
+        // For custom directories, skip wrealpath since file may not exist yet
+        if self.custom_directory.is_some() {
+            Ok(Some(path))
+        } else if let Some(canonicalized_path) = wrealpath(&path) {
             Ok(Some(canonicalized_path))
         } else {
             Err(std::io::Error::other(format!(
@@ -456,7 +460,7 @@ impl HistoryImpl {
                 continue;
             }
 
-            if !seen.insert(item.contents.to_owned()) {
+            if !seen.insert(item.contents.clone()) {
                 // This item was not inserted because it was already in the set, so delete the item at
                 // this index.
                 self.new_items.remove(idx);
@@ -511,23 +515,21 @@ impl HistoryImpl {
                 let Some(old_item) = local_file.decode_item(offset) else {
                     continue;
                 };
-
-                // If old item is newer than session always erase if in deleted.
-                if old_item.timestamp() > self.boundary_timestamp {
-                    if old_item.is_empty() || self.deleted_items.contains_key(old_item.str()) {
-                        continue;
-                    }
-                    lru.add_item(old_item);
-                } else {
-                    // If old item is older and in deleted items don't erase if added by
-                    // clear_session.
-                    if old_item.is_empty() || self.deleted_items.get(old_item.str()) == Some(&false)
-                    {
-                        continue;
-                    }
-                    // Add this old item.
-                    lru.add_item(old_item);
+                if old_item.is_empty() {
+                    continue;
                 }
+
+                // Check if this item should be deleted.
+                if let Some(&scope) = self.deleted_items.get(old_item.str()) {
+                    // If old item is newer than session always erase if in deleted.
+                    // If old item is older and in deleted items don't erase if added by clear_session.
+                    let delete = old_item.timestamp() > self.boundary_timestamp
+                        || scope == DeletionScope::AllSessions;
+                    if delete {
+                        continue;
+                    }
+                }
+                lru.add_item(old_item);
             }
         }
 
@@ -551,30 +553,12 @@ impl HistoryImpl {
         /// Default buffer size for flushing to the history file.
         const HISTORY_OUTPUT_BUFFER_SIZE: usize = 64 * 1024;
         // Write them out.
-        let mut err = None;
-        let mut buffer = Vec::with_capacity(HISTORY_OUTPUT_BUFFER_SIZE + 128);
+        let mut buffer = BufWriter::with_capacity(HISTORY_OUTPUT_BUFFER_SIZE + 128, dst);
         for item in items {
-            append_history_item_to_buffer(&item, &mut buffer);
-            if let Err(e) = flush_to_file(&mut buffer, dst, HISTORY_OUTPUT_BUFFER_SIZE) {
-                err = Some(e);
-                break;
-            }
-        }
-        if err.is_none() {
-            if let Err(e) = flush_to_file(&mut buffer, dst, 0) {
-                err = Some(e);
-            }
-        }
-        if let Some(err) = err {
-            flog!(
-                history_file,
-                "Error writing to temporary history file:",
-                err
-            );
-            Err(err)
-        } else {
-            Ok(())
+            item.write_to(&mut buffer)?;
         }
+        buffer.flush()?;
+        Ok(())
     }
 
     /// Saves history by rewriting the file.
@@ -587,7 +571,15 @@ impl HistoryImpl {
 
         let rewrite =
             |old_file: &File, tmp_file: &mut File| -> std::io::Result<PotentialUpdate<()>> {
-                self.rewrite_to_temporary_file(old_file, tmp_file)?;
+                let result = self.rewrite_to_temporary_file(old_file, tmp_file);
+                if let Err(err) = result {
+                    flog!(
+                        history_file,
+                        "Error writing to temporary history file:",
+                        err
+                    );
+                    return Err(err);
+                }
                 Ok(PotentialUpdate {
                     do_save: true,
                     data: (),
@@ -646,19 +638,20 @@ impl HistoryImpl {
         // So far so good. Write all items at or after first_unwritten_new_item_index. Note that we
         // write even a pending item - pending items are ignored by history within the command
         // itself, but should still be written to the file.
-        // Use a small buffer size for appending, we usually only have 1 item
+        // Use a small buffer size for appending, as we usually only have 1 item.
+        // Buffer everything and then write it all at once to avoid tearing writes (O_APPEND).
         let mut buffer = Vec::new();
         let mut new_first_index = self.first_unwritten_new_item_index;
         while new_first_index < self.new_items.len() {
             let item = &self.new_items[new_first_index];
             if item.should_write_to_disk() {
-                append_history_item_to_buffer(item, &mut buffer);
+                // Can't error writing to a buffer.
+                item.write_to(&mut buffer).unwrap();
             }
             // We wrote or skipped this item, hooray.
             new_first_index += 1;
         }
-
-        flush_to_file(&mut buffer, locked_history_file.get_mut(), 0)?;
+        locked_history_file.get_mut().write_all(&buffer)?;
         fsync(locked_history_file.get())?;
         self.first_unwritten_new_item_index = new_first_index;
 
@@ -714,7 +707,7 @@ impl HistoryImpl {
         if !ok {
             // We did not or could not append; rewrite the file ("vacuum" it).
             if let Err(e) = self.save_internal_via_rewrite(&history_path) {
-                flog!(history, "Rewriting history failed:", e)
+                flog!(history, "Rewriting history failed:", e);
             }
         }
     }
@@ -754,9 +747,10 @@ impl HistoryImpl {
         self.save(vacuum);
     }
 
-    fn new(name: WString) -> Self {
+    fn new(name: WString, custom_directory: Option<WString>) -> Self {
         Self {
             name,
+            custom_directory,
             new_items: vec![],
             first_unwritten_new_item_index: 0,
             has_pending_item: false,
@@ -808,7 +802,8 @@ impl HistoryImpl {
     /// Remove a history item.
     fn remove(&mut self, str_to_remove: &wstr) {
         // Add to our list of deleted items.
-        self.deleted_items.insert(str_to_remove.to_owned(), false);
+        self.deleted_items
+            .insert(str_to_remove.to_owned(), DeletionScope::AllSessions);
 
         for idx in (0..self.new_items.len()).rev() {
             let matched = self.new_items[idx].str() == str_to_remove;
@@ -833,7 +828,7 @@ impl HistoryImpl {
     /// Enable / disable automatic saving. Main thread only!
     fn disable_automatic_saving(&mut self) {
         self.disable_automatic_save_counter += 1;
-        assert!(self.disable_automatic_save_counter != 0); // overflow!
+        assert_ne!(self.disable_automatic_save_counter, 0); // overflow!
     }
 
     fn enable_automatic_saving(&mut self) {
@@ -848,7 +843,7 @@ impl HistoryImpl {
         self.first_unwritten_new_item_index = 0;
         self.file_contents = None;
         if let Ok(Some(filename)) = self.history_file_path() {
-            wunlink(&filename);
+            let _ = wunlink(&filename);
         }
         self.clear_file_state();
     }
@@ -856,7 +851,8 @@ impl HistoryImpl {
     /// Clears only session.
     fn clear_session(&mut self) {
         for item in &self.new_items {
-            self.deleted_items.insert(item.str().to_owned(), true);
+            self.deleted_items
+                .insert(item.str().to_owned(), DeletionScope::SessionOnly);
         }
 
         self.new_items.clear();
@@ -981,7 +977,7 @@ impl HistoryImpl {
             }
 
             if seen.insert(item.str().to_owned()) {
-                result.push(item.str().to_owned())
+                result.push(item.str().to_owned());
             }
         }
 
@@ -1119,13 +1115,12 @@ fn do_1_history_search(
 }
 
 /// Formats a single history record, including a trailing newline.
-///
-/// Returns nothing. The only possible failure involves formatting the timestamp. If that happens we
-/// simply omit the timestamp from the output.
 fn format_history_record(
     item: &HistoryItem,
     show_time_format: Option<&str>,
     null_terminate: bool,
+    parser: &Parser,
+    color_enabled: bool,
 ) -> WString {
     let mut result = WString::new();
     let seconds = time_to_seconds(item.timestamp());
@@ -1146,12 +1141,24 @@ fn format_history_record(
                 )
             } != 0
             {
-                result.push_utfstr(&cstr2wcstring(&timestamp_str[..]));
+                // SAFETY: strftime terminates the string with a null byte. If there is insufficient
+                // space, strftime returns 0.
+                let timestamp_cstr = CStr::from_bytes_until_nul(&timestamp_str).unwrap();
+                result.push_utfstr(&cstr2wcstring(timestamp_cstr));
             }
         }
     }
 
-    result.push_utfstr(item.str());
+    let mut command = item.str().to_owned();
+    if color_enabled {
+        command = bytes2wcstring(&highlight_and_colorize(
+            &command,
+            &parser.context(),
+            parser.vars(),
+        ));
+    }
+
+    result.push_utfstr(&command);
     result.push(if null_terminate { '\0' } else { '\n' });
     result
 }
@@ -1179,18 +1186,18 @@ fn should_import_bash_history_line(line: &wstr) -> bool {
     // Skip lines with [[...]] and ((...)) since we don't handle those constructs.
     // "<<" here is a proxy for heredocs (and herestrings).
     for seq in [L!("[["), L!("]]"), L!("(("), L!("))"), L!("<<")] {
-        if find_subslice(seq, line.as_char_slice()).is_some() {
+        if subslice_position(line.as_char_slice(), seq).is_some() {
             return false;
         }
     }
 
-    if ast::parse(line, ParseTreeFlags::empty(), None).errored() {
+    if ast::parse(line, ParseTreeFlags::default(), None).errored() {
         return false;
     }
 
     // In doing this test do not allow incomplete strings. Hence the "false" argument.
     let mut errors = Vec::new();
-    let _ = parse_util_detect_errors(line, Some(&mut errors), false);
+    let _ = detect_parse_errors(line, Some(&mut errors), false);
     errors.is_empty()
 }
 
@@ -1205,28 +1212,36 @@ impl History {
     /// call to resolve_pending. Any trailing ephemeral items are dropped.
     /// Exposed for testing.
     pub fn add(&self, item: HistoryItem, pending: bool) {
-        self.imp().add(item, pending, true)
+        self.imp().add(item, pending, true);
     }
 
     pub fn add_commandline(&self, s: WString) {
         let mut imp = self.imp();
         let when = imp.timestamp_now();
         let item = HistoryItem::new(s, when, PersistenceMode::Disk);
-        imp.add(item, false, true)
+        imp.add(item, false, true);
     }
 
-    pub fn new(name: &wstr) -> Arc<Self> {
-        Arc::new(Self(Mutex::new(HistoryImpl::new(name.to_owned()))))
+    /// Creates a new History with a custom directory path.
+    /// The history file will be stored at `{directory}/{name}_history`.
+    /// If the directory is None, it will be stored at path_get_data().
+    pub fn new(name: &wstr, directory: Option<WString>) -> Arc<Self> {
+        Arc::new(Self(Mutex::new(HistoryImpl::new(
+            name.to_owned(),
+            directory,
+        ))))
     }
 
-    /// Returns history with the given name, creating it if necessary.
+    /// Returns the history with the given name, creating it if necessary, using the default data directory.
+    /// This uses the HISTORIES global collection. Note it is possible to create a history without
+    /// placing it into this collection.
     pub fn with_name(name: &wstr) -> Arc<Self> {
         let mut histories = HISTORIES.lock().unwrap();
 
         if let Some(hist) = histories.get(name) {
             Arc::clone(hist)
         } else {
-            let hist = Self::new(name);
+            let hist = Self::new(name, None);
             histories.insert(name.to_owned(), Arc::clone(&hist));
             hist
         }
@@ -1244,12 +1259,12 @@ impl History {
 
     /// Remove a history item.
     pub fn remove(&self, s: &wstr) {
-        self.imp().remove(s)
+        self.imp().remove(s);
     }
 
     /// Remove any trailing ephemeral items.
     pub fn remove_ephemeral_items(&self) {
-        self.imp().remove_ephemeral_items()
+        self.imp().remove_ephemeral_items();
     }
 
     /// Add a new pending history item to the end, and then begin file detection on the items to
@@ -1269,7 +1284,7 @@ impl History {
 
         // Find all arguments that look like they could be file paths.
         let mut needs_sync_write = false;
-        let ast = ast::parse(s, ParseTreeFlags::empty(), None);
+        let ast = ast::parse(s, ParseTreeFlags::default(), None);
 
         let mut potential_paths = Vec::new();
         for node in ast.walk() {
@@ -1344,18 +1359,20 @@ impl History {
 
     /// Resolves any pending history items, so that they may be returned in history searches.
     pub fn resolve_pending(&self) {
-        self.imp().resolve_pending()
+        self.imp().resolve_pending();
     }
 
     /// Saves history.
     pub fn save(&self) {
-        self.imp().save(false)
+        self.imp().save(false);
     }
 
     /// Searches history.
     #[allow(clippy::too_many_arguments)]
     pub fn search(
         self: &Arc<Self>,
+        parser: &Parser,
+        streams: &mut IoStreams,
         search_type: SearchType,
         search_args: &[&wstr],
         show_time_format: Option<&str>,
@@ -1364,7 +1381,7 @@ impl History {
         null_terminate: bool,
         reverse: bool,
         cancel_check: &CancelChecker,
-        streams: &mut IoStreams,
+        color_enabled: bool,
     ) -> bool {
         let mut remaining = max_items;
         let mut collected = Vec::new();
@@ -1376,7 +1393,14 @@ impl History {
                 return ControlFlow::Break(());
             }
             remaining -= 1;
-            let formatted_record = format_history_record(item, show_time_format, null_terminate);
+            let formatted_record = format_history_record(
+                item,
+                show_time_format,
+                null_terminate,
+                parser,
+                color_enabled,
+            );
+
             if reverse {
                 // We need to collect this for later.
                 collected.push(formatted_record);
@@ -1395,9 +1419,9 @@ impl History {
             // The user had no search terms; just append everything.
             do_1_history_search(
                 Arc::clone(self),
-                SearchType::MatchEverything,
+                SearchType::Contains,
                 WString::new(),
-                false,
+                true,
                 &mut func,
                 cancel_check,
             );
@@ -1441,27 +1465,27 @@ impl History {
 
     /// Irreversibly clears history.
     pub fn clear(&self) {
-        self.imp().clear()
+        self.imp().clear();
     }
 
     /// Irreversibly clears history for the current session.
     pub fn clear_session(&self) {
-        self.imp().clear_session()
+        self.imp().clear_session();
     }
 
     /// Populates from older location (in config path, rather than data path).
     pub fn populate_from_config_path(&self) {
-        self.imp().populate_from_config_path()
+        self.imp().populate_from_config_path();
     }
 
     /// Populates from a bash history file.
     pub fn populate_from_bash<R: BufRead>(&self, contents: R) {
-        self.imp().populate_from_bash(contents)
+        self.imp().populate_from_bash(contents);
     }
 
     /// Incorporates the history of other shells into this history.
     pub fn incorporate_external_changes(&self) {
-        self.imp().incorporate_external_changes()
+        self.imp().incorporate_external_changes();
     }
 
     /// Gets all the history into a list. This is intended for the $history environment variable.
@@ -1525,13 +1549,16 @@ pub struct HistorySearch {
 }
 
 impl HistorySearch {
-    pub fn new(hist: Arc<History>, s: WString) -> Self {
+    #[cfg(test)]
+    fn new(hist: Arc<History>, s: WString) -> Self {
         Self::new_with(hist, s, SearchType::Contains, SearchFlags::default(), 0)
     }
-    pub fn new_with_type(hist: Arc<History>, s: WString, search_type: SearchType) -> Self {
+    #[cfg(test)]
+    fn new_with_type(hist: Arc<History>, s: WString, search_type: SearchType) -> Self {
         Self::new_with(hist, s, search_type, SearchFlags::default(), 0)
     }
-    pub fn new_with_flags(hist: Arc<History>, s: WString, flags: SearchFlags) -> Self {
+    #[cfg(test)]
+    fn new_with_flags(hist: Arc<History>, s: WString, flags: SearchFlags) -> Self {
         Self::new_with(hist, s, SearchType::Contains, flags, 0)
     }
     /// Constructs a new history search.
@@ -1566,7 +1593,7 @@ impl HistorySearch {
     }
 
     pub fn prepare_to_search_after_deletion(&mut self) {
-        assert!(self.current_index != 0);
+        assert_ne!(self.current_index, 0);
         self.current_index -= 1;
         self.current_item = None;
     }
@@ -1588,7 +1615,7 @@ impl HistorySearch {
             match direction {
                 SearchDirection::Backward => index += 1,
                 SearchDirection::Forward => index -= 1,
-            };
+            }
 
             if self.current_index == invalid_index {
                 return false;
@@ -1636,6 +1663,10 @@ impl HistorySearch {
         self.current_item.as_ref().expect("No current item")
     }
 
+    pub fn canon_term(&self) -> &wstr {
+        &self.canon_term
+    }
+
     /// Returns the current search result item contents.
     ///
     /// # Panics
@@ -1733,13 +1764,12 @@ pub fn expand_and_detect_paths<P: IntoIterator<Item = WString>>(
 /// Given a list of proposed paths and a context, expand each one and see if it refers to a file.
 /// Wildcard expansions are suppressed.
 /// Returns `true` if `paths` is empty or every path is valid.
-pub fn all_paths_are_valid<P: IntoIterator<Item = WString>>(
-    paths: P,
-    ctx: &OperationContext<'_>,
-) -> bool {
+pub fn all_paths_are_valid(paths: &[WString], ctx: &OperationContext<'_>) -> bool {
     assert_is_background_thread();
     let working_directory = ctx.vars().get_pwd_slash();
-    for mut path in paths {
+    let mut path = WString::new();
+    for unexpanded_path in paths {
+        path.clone_from(unexpanded_path);
         if ctx.check_cancel() {
             return false;
         }
@@ -1776,23 +1806,19 @@ mod tests {
         History, HistoryItem, HistorySearch, PathList, PersistenceMode, SearchDirection,
         SearchFlags, SearchType, VACUUM_FREQUENCY,
     };
-    use crate::common::ESCAPE_TEST_CHAR;
-    use crate::common::{ScopeGuard, bytes2wcstring, wcs2bytes, wcs2osstring};
+    use crate::common::{ESCAPE_TEST_CHAR, osstr2wcstring, wcs2bytes};
     use crate::env::{EnvMode, EnvSetMode, EnvStack};
     use crate::fs::{LockedFile, WriteMethod};
-    use crate::path::path_get_data;
     use crate::prelude::*;
-    use crate::tests::prelude::*;
-    use crate::wcstringutil::{string_prefixes_string, string_prefixes_string_case_insensitive};
     use fish_build_helper::workspace_root;
-    use rand::Rng;
+    use fish_wcstringutil::{string_prefixes_string, string_prefixes_string_case_insensitive};
+    use rand::Rng as _;
     use rand::rngs::ThreadRng;
     use std::collections::VecDeque;
+    use std::ffi::OsString;
     use std::io::BufReader;
-    use std::os::unix::ffi::OsStrExt;
     use std::sync::Arc;
-    use std::time::UNIX_EPOCH;
-    use std::time::{Duration, SystemTime};
+    use std::time::{Duration, SystemTime, UNIX_EPOCH};
 
     fn history_contains(history: &History, txt: &wstr) -> bool {
         for i in 1.. {
@@ -1808,6 +1834,11 @@ mod tests {
         false
     }
 
+    // Helper to create a history with a custom directory, for testing.
+    fn create_test_history(name: &wstr, custom_dir: &wstr) -> Arc<History> {
+        History::new(name, Some(custom_dir.to_owned()))
+    }
+
     fn random_string(rng: &mut ThreadRng) -> WString {
         let mut result = WString::new();
         let max = rng.random_range(1..=32);
@@ -1821,9 +1852,10 @@ mod tests {
     }
 
     #[test]
-    #[serial]
     fn test_history() {
-        let _cleanup = test_init();
+        let tmpdir = fish_tempfile::new_dir().unwrap();
+        let hist_dir = osstr2wcstring(tmpdir.path());
+
         macro_rules! test_history_matches {
             ($search:expr, $expected:expr) => {
                 let expected: Vec<&wstr> = $expected;
@@ -1850,7 +1882,7 @@ mod tests {
         let nocase = SearchFlags::IGNORE_CASE;
 
         // Populate a history.
-        let history = History::with_name(L!("test_history"));
+        let history = create_test_history(L!("test_history"), &hist_dir);
         history.clear();
         for s in items {
             history.add_commandline(s.to_owned());
@@ -1933,7 +1965,7 @@ mod tests {
         let mut rng = rand::rng();
         for i in 1..=max {
             // Generate a value.
-            let mut value = WString::from_str("test item ") + &i.to_wstring()[..];
+            let mut value = L!("test item ").to_owned() + &i.to_wstring()[..];
 
             // Maybe add some backslashes.
             if i % 3 == 0 {
@@ -2002,9 +2034,9 @@ mod tests {
         result
     }
 
-    fn pound_on_history(item_count: usize, idx: usize) -> Arc<History> {
+    fn write_history_entries(dir: &wstr, item_count: usize, idx: usize) -> Arc<History> {
         // Called in child thread to modify history.
-        let hist = History::new(L!("race_test"));
+        let hist = create_test_history(L!("race_test"), dir);
         let hist_lines = generate_history_lines(item_count, idx);
         for line in hist_lines {
             hist.add_commandline(line);
@@ -2014,23 +2046,16 @@ mod tests {
     }
 
     #[test]
-    #[serial]
     fn test_history_races() {
-        let _cleanup = test_init();
+        // Place history in a temp directory.
+        let tmpdir = fish_tempfile::new_dir().unwrap();
+        let hist_dir = osstr2wcstring(tmpdir.path());
 
-        let tmp_path = std::env::current_dir()
-            .unwrap()
-            .join("history-races-test-balloon");
-        std::fs::write(&tmp_path, []).unwrap();
-        let _cleanup = ScopeGuard::new((), |()| {
-            std::fs::remove_file(&tmp_path).unwrap();
-        });
-        if LockedFile::new(
-            crate::fs::LockingMode::Exclusive(WriteMethod::RenameIntoPlace),
-            &bytes2wcstring(tmp_path.as_os_str().as_bytes()),
-        )
-        .is_err()
-        {
+        // Skip tests if we can't get an exclusive lock on a file in that directory.
+        let tmp_balloon = tmpdir.path().join("history-races-test-balloon");
+        std::fs::write(&tmp_balloon, []).unwrap();
+        let mode = crate::fs::LockingMode::Exclusive(WriteMethod::RenameIntoPlace);
+        if LockedFile::new(mode, &osstr2wcstring(&tmp_balloon)).is_err() {
             return;
         }
 
@@ -2044,12 +2069,13 @@ mod tests {
         const ITEM_COUNT: usize = 256;
 
         // Ensure history is clear.
-        History::new(L!("race_test")).clear();
+        create_test_history(L!("race_test"), &hist_dir).clear();
 
         let mut children = Vec::with_capacity(RACE_COUNT);
         for i in 0..RACE_COUNT {
+            let hist_dir = hist_dir.clone();
             children.push(std::thread::spawn(move || {
-                pound_on_history(ITEM_COUNT, i);
+                write_history_entries(&hist_dir, ITEM_COUNT, i);
             }));
         }
 
@@ -2066,7 +2092,7 @@ mod tests {
         time_barrier();
 
         // Ensure that we got sane, sorted results.
-        let hist = History::new(L!("race_test"));
+        let hist = create_test_history(L!("race_test"), &hist_dir);
 
         // History is enumerated from most recent to least
         // Every item should be the last item in some array
@@ -2103,7 +2129,7 @@ mod tests {
                 );
                 for list in &expected_lines {
                     if !list.is_empty() {
-                        printf!("\tRemaining: %s\n", list.last().unwrap())
+                        printf!("\tRemaining: %s\n", list.last().unwrap());
                     }
                 }
             }
@@ -2120,25 +2146,26 @@ mod tests {
     }
 
     #[test]
-    #[serial]
     fn test_history_external_rewrites() {
-        let _cleanup = test_init();
+        // Place history in a temp directory.
+        let tmpdir = fish_tempfile::new_dir().unwrap();
+        let hist_dir = osstr2wcstring(tmpdir.path());
 
         // Write some history to disk.
         {
-            let hist = pound_on_history(VACUUM_FREQUENCY / 2, 0);
+            let hist = write_history_entries(&hist_dir, VACUUM_FREQUENCY / 2, 0);
             hist.add_commandline("needle".into());
             hist.save();
         }
         std::thread::sleep(Duration::from_secs(1));
 
         // Read history from disk.
-        let hist = History::new(L!("race_test"));
+        let hist = create_test_history(L!("race_test"), &hist_dir);
         assert_eq!(hist.item_at_index(1).unwrap().str(), "needle");
 
         // Add items until we rewrite the file.
         // In practice this might be done by another shell.
-        pound_on_history(VACUUM_FREQUENCY, 0);
+        write_history_entries(&hist_dir, VACUUM_FREQUENCY, 0);
 
         for i in 1.. {
             if hist.item_at_index(i).unwrap().str() == "needle" {
@@ -2148,16 +2175,21 @@ mod tests {
     }
 
     #[test]
-    #[serial]
     fn test_history_merge() {
-        let _cleanup = test_init();
+        let tmpdir = fish_tempfile::new_dir().unwrap();
+        let hist_dir = osstr2wcstring(tmpdir.path());
+
         // In a single fish process, only one history is allowed to exist with the given name But it's
         // common to have multiple history instances with the same name active in different processes,
         // e.g. when you have multiple shells open. We try to get that right and merge all their history
         // together. Test that case.
         const COUNT: usize = 3;
         let name = L!("merge_test");
-        let hists = [History::new(name), History::new(name), History::new(name)];
+        let hists = [
+            create_test_history(name, &hist_dir),
+            create_test_history(name, &hist_dir),
+            create_test_history(name, &hist_dir),
+        ];
         let texts = [L!("History 1"), L!("History 2"), L!("History 3")];
         let alt_texts = [
             L!("History Alt 1"),
@@ -2180,7 +2212,7 @@ mod tests {
 
         // Save them.
         for hist in &hists {
-            hist.save()
+            hist.save();
         }
 
         // Make sure each history contains what it ought to, but they have not leaked into each other.
@@ -2196,7 +2228,7 @@ mod tests {
         // Make a new history. It should contain everything. The time_barrier() is so that the timestamp
         // is newer, since we only pick up items whose timestamp is before the birth stamp.
         time_barrier();
-        let everything = History::new(name);
+        let everything = create_test_history(name, &hist_dir);
         for text in texts {
             assert!(history_contains(&everything, text));
         }
@@ -2257,10 +2289,13 @@ mod tests {
     }
 
     #[test]
-    #[serial]
     fn test_history_path_detection() {
-        let _cleanup = test_init();
         // Regression test for #7582.
+        // Temporary directory for the history files.
+        let hist_tmpdir = fish_tempfile::new_dir().unwrap();
+        let hist_dir = Some(osstr2wcstring(hist_tmpdir.path()));
+
+        // Temporary directory for the files we will detect.
         let tmpdir = fish_tempfile::new_dir().unwrap();
 
         // Place one valid file in the directory.
@@ -2275,7 +2310,7 @@ mod tests {
         test_vars.set_one(L!("PWD"), global_mode, wdir_path.clone());
         test_vars.set_one(L!("HOME"), global_mode, wdir_path.clone());
 
-        let history = History::with_name(L!("path_detection"));
+        let history = History::new(L!("path_detection"), hist_dir);
         history.clear();
         assert_eq!(history.size(), 0);
         history.add_pending_with_file_detection(
@@ -2361,30 +2396,31 @@ mod tests {
         history.clear();
     }
 
-    fn install_sample_history(name: &wstr) {
-        let path = path_get_data().expect("Failed to get data directory");
+    fn install_sample_history(name: &wstr, hist_dir: &wstr) {
+        let dst_hist_path: OsString = format!("{}/{}_history", hist_dir, name).into();
         std::fs::copy(
             workspace_root()
                 .join("tests")
                 .join(std::str::from_utf8(&wcs2bytes(name)).unwrap()),
-            wcs2osstring(&(path + L!("/") + name + L!("_history"))),
+            dst_hist_path,
         )
         .unwrap();
     }
 
     #[test]
-    #[serial]
     fn test_history_formats() {
-        let _cleanup = test_init();
+        let tmpdir = fish_tempfile::new_dir().unwrap();
+        let hist_dir = osstr2wcstring(tmpdir.path());
+
         // Test inferring and reading legacy and bash history formats.
         let name = L!("history_sample_fish_2_0");
-        install_sample_history(name);
+        install_sample_history(name, &hist_dir);
         let expected: Vec<WString> = vec![
             "echo this has\\\nbackslashes".into(),
             "function foo\necho bar\nend".into(),
             "echo alpha".into(),
         ];
-        let test_history_imported = History::with_name(name);
+        let test_history_imported = create_test_history(name, &hist_dir);
         assert_eq!(test_history_imported.get_history(), expected);
         test_history_imported.clear();
 
@@ -2403,16 +2439,16 @@ mod tests {
             "history --help".into(),
             "echo foo".into(),
         ];
-        let test_history_imported_from_bash = History::with_name(L!("bash_import"));
+        let test_history_imported_from_bash = create_test_history(L!("bash_import"), &hist_dir);
         let file = std::fs::File::open(workspace_root().join("tests/history_sample_bash")).unwrap();
         test_history_imported_from_bash.populate_from_bash(BufReader::new(file));
         assert_eq!(test_history_imported_from_bash.get_history(), expected);
         test_history_imported_from_bash.clear();
 
         let name = L!("history_sample_corrupt1");
-        install_sample_history(name);
+        install_sample_history(name, &hist_dir);
         // We simply invoke get_string_representation. If we don't die, the test is a success.
-        let test_history_imported_from_corrupted = History::with_name(name);
+        let test_history_imported_from_corrupted = create_test_history(name, &hist_dir);
         let expected: Vec<WString> = vec![
             "no_newline_at_end_of_file".into(),
             "corrupt_prefix".into(),
